__SYCL_PARAM_TRAITS_SPEC(platform, profile, string_class)
__SYCL_PARAM_TRAITS_SPEC(platform, version, string_class)
__SYCL_PARAM_TRAITS_SPEC(platform, name, string_class)
__SYCL_PARAM_TRAITS_SPEC(platform, vendor, string_class)
__SYCL_PARAM_TRAITS_SPEC(platform, extensions, vector_class<string_class>)
