//==----------- fpga_extensions.hpp --- SYCL FPGA Extensions ---------------==//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#pragma once
#include <CL/sycl/INTEL/fpga_device_selector.hpp>
#include <CL/sycl/INTEL/fpga_lsu.hpp>
#include <CL/sycl/INTEL/fpga_reg.hpp>
#include <CL/sycl/INTEL/pipes.hpp>
