#include <CL/sycl/properties/accessor_properties.hpp>
#include <CL/sycl/properties/buffer_properties.hpp>
#include <CL/sycl/properties/context_properties.hpp>
#include <CL/sycl/properties/image_properties.hpp>
#include <CL/sycl/properties/queue_properties.hpp>
#include <CL/sycl/properties/reduction_properties.hpp>
