/* file: math_common_undefine.h */

/*
** Copyright  (C) 1985-2018 Intel Corporation. All rights reserved.
**
** The information and source code contained herein is the exclusive property
** of Intel Corporation and may not be disclosed, examined, or reproduced in
** whole or in part without explicit written authorization from the Company.
**
*/

#undef LIBIMF_EXTERN_C
#undef _LIBIMF_FORCEINLINE
#undef __IMFLONGDOUBLE
#undef _LIBIMF_CPP_EXCEPT_SPEC

#undef _LIBIMF_PUBAPI
#undef _LIBIMF_PUBAPI_INL
#undef _LIBIMF_PUBVAR

#undef _LIBIMF_DBL_XDBL
