__SYCL_PARAM_TRAITS_SPEC(sycl::property::buffer::use_host_ptr)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::buffer::use_mutex)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::buffer::context_bound)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::image::use_host_ptr)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::image::use_mutex)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::image::context_bound)
__SYCL_PARAM_TRAITS_SPEC(sycl::ext::oneapi::property::buffer::use_pinned_host_memory)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::noinit)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::context::cuda::use_primary_context)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::queue::in_order)
__SYCL_PARAM_TRAITS_SPEC(sycl::property::reduction::initialize_to_identity)
