//==------------ pi.def Plugin Interface list of API -----------------------==//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef _PI_API
#error Undefined _PI_API macro expansion
#endif

// The list of all PI interfaces wrapped with _PI_API macro.
// This is for convinience of doing same thing for all interfaces, e.g.
// declare, define, initialize.
//
// Platform
_PI_API(piPlatformsGet)
_PI_API(piPlatformGetInfo)
_PI_API(piextPlatformGetNativeHandle)
_PI_API(piextPlatformCreateWithNativeHandle)
// Device
_PI_API(piDevicesGet)
_PI_API(piDeviceGetInfo)
_PI_API(piDevicePartition)
_PI_API(piDeviceRetain)
_PI_API(piDeviceRelease)
_PI_API(piextDeviceSelectBinary)
_PI_API(piextGetDeviceFunctionPointer)
_PI_API(piextDeviceGetNativeHandle)
_PI_API(piextDeviceCreateWithNativeHandle)
// Context
_PI_API(piContextCreate)
_PI_API(piContextGetInfo)
_PI_API(piContextRetain)
_PI_API(piContextRelease)
_PI_API(piextContextSetExtendedDeleter)
_PI_API(piextContextGetNativeHandle)
_PI_API(piextContextCreateWithNativeHandle)
// Queue
_PI_API(piQueueCreate)
_PI_API(piQueueGetInfo)
_PI_API(piQueueFinish)
_PI_API(piQueueRetain)
_PI_API(piQueueRelease)
_PI_API(piextQueueGetNativeHandle)
_PI_API(piextQueueCreateWithNativeHandle)
// Memory
_PI_API(piMemBufferCreate)
_PI_API(piMemImageCreate)
_PI_API(piMemGetInfo)
_PI_API(piMemImageGetInfo)
_PI_API(piMemRetain)
_PI_API(piMemRelease)
_PI_API(piMemBufferPartition)
_PI_API(piextMemGetNativeHandle)
_PI_API(piextMemCreateWithNativeHandle)
// Program
_PI_API(piProgramCreate)
_PI_API(piclProgramCreateWithSource)
_PI_API(piProgramCreateWithBinary)
_PI_API(piProgramGetInfo)
_PI_API(piProgramCompile)
_PI_API(piProgramBuild)
_PI_API(piProgramLink)
_PI_API(piProgramGetBuildInfo)
_PI_API(piProgramRetain)
_PI_API(piProgramRelease)
_PI_API(piextProgramSetSpecializationConstant)
_PI_API(piextProgramGetNativeHandle)
_PI_API(piextProgramCreateWithNativeHandle)
// Kernel
_PI_API(piKernelCreate)
_PI_API(piKernelSetArg)
_PI_API(piKernelGetInfo)
_PI_API(piKernelGetGroupInfo)
_PI_API(piKernelGetSubGroupInfo)
_PI_API(piKernelRetain)
_PI_API(piKernelRelease)
_PI_API(piextKernelSetArgPointer)
_PI_API(piKernelSetExecInfo)
_PI_API(piextKernelCreateWithNativeHandle)
_PI_API(piextKernelGetNativeHandle)
// Event
_PI_API(piEventCreate)
_PI_API(piEventGetInfo)
_PI_API(piEventGetProfilingInfo)
_PI_API(piEventsWait)
_PI_API(piEventSetCallback)
_PI_API(piEventSetStatus)
_PI_API(piEventRetain)
_PI_API(piEventRelease)
_PI_API(piextEventGetNativeHandle)
_PI_API(piextEventCreateWithNativeHandle)
// Sampler
_PI_API(piSamplerCreate)
_PI_API(piSamplerGetInfo)
_PI_API(piSamplerRetain)
_PI_API(piSamplerRelease)
// Queue commands
_PI_API(piEnqueueKernelLaunch)
_PI_API(piEnqueueNativeKernel)
_PI_API(piEnqueueEventsWait)
_PI_API(piEnqueueEventsWaitWithBarrier)
_PI_API(piEnqueueMemBufferRead)
_PI_API(piEnqueueMemBufferReadRect)
_PI_API(piEnqueueMemBufferWrite)
_PI_API(piEnqueueMemBufferWriteRect)
_PI_API(piEnqueueMemBufferCopy)
_PI_API(piEnqueueMemBufferCopyRect)
_PI_API(piEnqueueMemBufferFill)
_PI_API(piEnqueueMemImageRead)
_PI_API(piEnqueueMemImageWrite)
_PI_API(piEnqueueMemImageCopy)
_PI_API(piEnqueueMemImageFill)
_PI_API(piEnqueueMemBufferMap)
_PI_API(piEnqueueMemUnmap)
// USM
_PI_API(piextUSMHostAlloc)
_PI_API(piextUSMDeviceAlloc)
_PI_API(piextUSMSharedAlloc)
_PI_API(piextUSMFree)
_PI_API(piextUSMEnqueueMemset)
_PI_API(piextUSMEnqueueMemcpy)
_PI_API(piextUSMEnqueuePrefetch)
_PI_API(piextUSMEnqueueMemAdvise)
_PI_API(piextUSMGetMemAllocInfo)

_PI_API(piextKernelSetArgMemObj)
_PI_API(piextKernelSetArgSampler)
_PI_API(piTearDown)

#undef _PI_API
