//==--------------- atomic.hpp - SYCL_ONEAPI_extended_atomics --------------==//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#pragma once

#include <CL/sycl/ONEAPI/atomic_accessor.hpp>
#include <CL/sycl/ONEAPI/atomic_enums.hpp>
#include <CL/sycl/ONEAPI/atomic_fence.hpp>
#include <CL/sycl/ONEAPI/atomic_ref.hpp>
